source("famtoo.q")
source("auxpsglm.q")

##############################################################
##                                                           #
## psglm Version 1.0: 07/01/97  by Richard P. Waterman       # 
##                                                           #
##  S-Plus function to implement projected score methodology #
##  to obtain approximate conditional estimates for          #
##  independent strata Generalized Linear Models with        #
##  canonical link functions.                                #
##                                                           #
##  Permission is hereby given to StatLib to redistribute    #
##  this software. The software can be freely used for       #
##  non-commercial purposes, and can be freely distributed   #
##  for non-commercial purposes only.                        #
##  The copyright is retained by the developer.              #
##  Copyright 1997 Richard P. Waterman                       #
##                                                           # 
##  Please mail comments, suggestions and errors to          #
##  Richard Waterman waterman@compstat.wharton.upenn.edu     #
##                                                           #
##############################################################


psglm <- function(nu = matrix(1,nrow=dim(y)[1], ncol=dim(y)[2]),
             y,x,family=gaussian, weights = rep(1, length(y[, 1])),
             start=NULL,expinfo = T,conditional=T, paramtest = NULL,
             control = glm.control(epsilon=1e-4,maxit=100,...),...)

{
    call <- match.call()
    stepsize <- sqrt(control$epsilon)
    convcrit <- control$epsilon
    maxit    <- control$maxit
    trace <- control$trace


    if((conditional == F) && (expinfo == F)) stop("Crazy option
    combination .. see help page. If conditional == F then expinfo must
    equal T")

#   nu is a parameter that gets fed to the cumulant function.
#   it will be trials for  a binomial.
#   Can't subsume trials into the weights like in glm because
#   of non-linearities at this higher order.


#   Check that y is a matrix.

    if(!is.matrix(y)) stop(paste("y must be a matrix"))

#   Is x as a matrix or an array?

    if(!( is.matrix(x) || is.array(x))) stop(paste("x must be a matrix or
    an array"))

    if( is.matrix(x) & (ncol(y) != nrow(x))) {
    	stop(paste("Columns of y must equal rows of x")) 
    }


#   How much data is there?

    rows <- nrow(y)
    cols <- ncol(y)


#   This expands the design matrix into a design array.
#   Right for the rasch model etc, something in which the design matrix is
#   the same for every row.

    if(is.matrix(x)){
    	x<- array(rep(x,rep(rows,length(x))),c(rows,cols,ncol(x)),
    	dimnames=list(NULL,NULL,dimnames(x)[[2]]))
    }	    	    	    	    	    	    	          
    	    	
#   How many parameters of interest?
    params <- dim(x)[3]


#   Check the weights
    if(!length(weights))
	weights <- matrix(rep(1, rows*cols),ncol=cols)
    else {
        if(any(weights < 0))
	    stop("negative weights not allowed")
        else 
    	    if(length(weights) != rows)
 		stop("length of weights must equal the number of rows
    	    	    of y")
    }

# If the family is Binomial check that the trials match

    if(familytoo(family)$family[1] == "Binomial") {
    	if(!is.matrix(nu)) stop("When it\'s binomial trials must be a matrix")
    	if(sum((dim(nu) - dim(y))^2 ) != 0) 
    	    stop("Dimensions of trials and y must match")
    } 

    if(trace) cat("\n INITIALIZING")

#   Make some starting values if they are not supplied.
#   This can take rather a long time if there are many strata.
#   Collapse y over rows or columns for the initial estimates.

    if(!length(start)){
    	psi <- 
    	(if(familytoo(family)$family[1] != "Gamma") 1 else -1) * 
    	glm(apply(y*weights,2,mean)/apply(nu*weights,2,mean)~apply(x,c(2,3),
    	    mean),family=family,
    	    control=glm.control(maxit=1))$coefficients[2:(params + 1)]

    	lambda <- (if(familytoo(family)$family[1] != "Gamma") 1 else -1) * 
    	familytoo(family)$link(apply(y,1,mean)/apply(nu,1,mean))
    }
    else {
    	psi <- start[[1]]
    	lambda <- start[[2]]
    }

    if(trace) cat("\n FINISHED INITIALIZING")


    if(trace){
    	cat("\n PARAMETERS OF INTEREST",length(psi))
    	cat("\n STRATA",length(lambda))
    	cat("\n DIM Y ",dim(y))
    	cat("\n DIM X ",dim(x))
    	cat("\n PSI START VALUES",psi)
    	cat("\n ENTERING MAIN LOOP\n")
    }

#   Containers to keep track of the parameter estimates.

    psireg <- matrix(rep(1,params * 2),ncol=2)
    lambdareg <- matrix(rep(1,rows * 2),ncol=2)

#   A counter and flag for the loop.

    itcount <- 1 
    iterkeep <- T



    while((iterkeep==T)){

#   Find the cumulant functions.

        cumlist <- cumcalc(x,psi,lambda,family,nu)

        if(trace){
    	    cat("\nSCORE",
            utwoa(x,y,weights,cumlist[[1]],cumlist[[2]],
                              cumlist[[3]],cumlist[[4]])) 
    	}

    	if(expinfo == T){
    	    if(conditional){
    	    	psi<-c(newpsival(
            	expinfoa(x,weights,cumlist[[2]]),
            	utwoa(x,y,weights,cumlist[[1]],
                      cumlist[[2]],cumlist[[3]],cumlist[[4]]),
            	psi))
    	    }
    		else{
    	        psi<-c(newpsival(
            	expinfoa(x,weights,cumlist[[2]]),
            	uonea(x,y,weights,cumlist[[1]]),psi))
    	    	}
    	    }
    	else {
    	    info <- -effinfo(y,x,family,weights,psi,lambda,stepsize,nu)
    	    psi<-c(newpsival(info,
    	    utwoa(x,y,weights,
            cumlist[[1]],cumlist[[2]],cumlist[[3]],cumlist[[4]]),psi))
    	}

    	cumlist <- cumcalc(x,psi,lambda,family,nu)
    	lambda <- newlambdaval(apply(cumlist[[2]]*weights,1,sum),
                               apply((y - cumlist[[1]])*weights,1,sum),lambda)



######################################
##### Check for convergence
######################################

#   Update the containers
    	psireg[,((itcount - 1)%%2 + 1)] <- psi  
    	lambdareg[,((itcount - 1)%%2 + 1)] <- lambda
    	if((max(apply(psireg,c(1),convergecheck),na.rm=T) < convcrit) &&
      	(max(apply(lambdareg,c(1),convergecheck),na.rm=T) < convcrit))
    	    {iterkeep <- F}

    	if(itcount == maxit) {
    	    iterkeep <- F 
    	    stop("maximum iterations reached")
    	}


    	if(trace){
    	    cat("\n CONVERGENCE DIAGNOSTICS")
    	    cat("\n\t PSI", psi)
    	    cat("\n\t MAX PSI RATIO ",
              	max(apply(psireg,c(1),convergecheck),na.rm=T))
    	    cat("\n\t MAX LAMBDA RATIO", 
            	max(apply(lambdareg,c(1),convergecheck),na.rm=T))
    	    cat("\n ITERATION",itcount,"\n")
    	}

    	itcount <- itcount + 1
    }
    if(trace) cat("\n EXITING MAIN LOOP\n")
#   Now out of the main loop.
#   Save a copy of psi in coef.
    coef <- psi 

#   Not sure why I don't put this in the summary function!
    if(length(dimnames(x)[[3]]) != params){
    	names(coef) <- paste("b", 1:length(coef), sep= "")}
    else {names(coef) <- dimnames(x)[[3]]}

#   Decide which information to return.
    if(conditional){
    	info <- effinfo(y,x,family,weights,psi,lambda,stepsize,nu)
    }
    else {
    	info <- effinfouone(x,weights,cumlist[[2]])
    }



####################################################
################### Score test section #############
####################################################

    teststat <- NULL

    itcounttwo <- 1 
    iterkeep <- T
#   only do this if some hypotheses stated.

    if(length(paramtest)>0){

    	if(trace) cat("\n STARTING SCORE TEST\n")

#   Construct the new x array, excluding the columns we are
#   testing. But first check to see if we are testing them all.

    	if(sum(paramtest) == (params*(params+1)/2)){
    	    testall <- T
    	}
    	else{
    	    testall <- F
    	}

#   Now exclude if necessary.

    	if(testall == F){
    	    xrestrict <- x[,,-paramtest,drop=F]
    	    psi <- psi[-paramtest]
    	} 
    	else{
    	    xrestrict <- x
    	    psi <- rep(0,params)
    	}

    	if(testall == F){
    	    psireg <- matrix(rep(1,length(psi) * 2),ncol=2)
    	}
    	else {
    	    psireg <- NULL
    	}

#   lambdareg <- matrix(rep(1,rows * 2),ncol=2)


    	while((iterkeep == T)){

    	    if(itcounttwo == maxit) iterkeep <- F 

#   Find the cell means, fed to the cumulant functions.

    	    if(testall == F){
    	    	cumlist <- cumcalc(xrestrict,psi,lambda,family,nu)

    	    	if(expinfo == T){
    	    	    if(conditional){
    	    	    	psi<-c(newpsival(
            	    	expinfoa(xrestrict,weights,cumlist[[2]]),
            	    	utwoa(xrestrict,y,weights,
                        cumlist[[1]],cumlist[[2]],cumlist[[3]],cumlist[[4]]),
                         psi))
    	    	    }
    	    	    else{
    	    	    	psi<-c(newpsival(
            	    	expinfoa(xrestrict,weights,cumlist[[2]]),
            	    	uonea(xrestrict,y,weights,cumlist[[1]]), psi))
    	    	    }
    	    	}
    	    else {
    	    	infoscore <- -effinfo(y,xrestrict,family,weights,psi,lambda,
                stepsize,nu) 
    	    	psi<-c(newpsival(infoscore,
    	    	utwoa(xrestrict,y,weights,cumlist[[1]],cumlist[[2]],
                                  cumlist[[3]],cumlist[[4]]),psi))
    	    }
    	}

    	cumlist <- cumcalc(xrestrict,psi,lambda,family,nu)
    	lambda <- newlambdaval(apply(cumlist[[2]]*weights,1,sum),
                       apply((y - cumlist[[1]])*weights,1,sum),lambda)




##############################################
##### Check for convergence in score test  ###
##############################################

    	lambdareg[,((itcounttwo - 1)%%2 + 1)] <- lambda
    	if(testall == F){
    	    psireg[,((itcounttwo - 1)%%2 + 1)] <- psi  
    	    if((max(apply(psireg,c(1),convergecheck),na.rm=T) < convcrit) &&
      	    (max(apply(lambdareg,c(1),convergecheck),na.rm=T) < convcrit))
    	    {
    	    	iterkeep <- F
    	    }
    	}
    	else {
    	    if(max(apply(lambdareg,c(1),convergecheck),na.rm=T) < convcrit)
    	    {
    	    	iterkeep <- F
    	    }
    	}

    	if(trace){
    	    cat("\n SCORE TEST CONVERGENCE DIAGNOSTICS")
    	    cat("\n\t PSI", psi)
    	    if(testall == F) {
    	    	cat("\n\t MAX PSI RATIO ",
                max(apply(psireg,c(1),convergecheck),na.rm=T)) 
            }
    	    cat("\n\t MAX LAMBDA RATIO", 
            max(apply(lambdareg,c(1),convergecheck),na.rm=T))
    	    cat("\n ITERATION",itcounttwo,"\n")
    	}

    	if(itcounttwo == maxit) {
    	    iterkeep <- F 
    	    stop("maximum iterations for score test reached")
    	}

    	itcounttwo <- itcounttwo + 1
    }

    if(trace) cat("\nEXITING SCORE TEST\n")

#   reconstruct psi if testall == F

    if(testall == F){
    	psi[seq(params)[-paramtest]] <- psi
    	psi[paramtest] <- 0
    }

#   now decide which sort of score function to return

    if(conditional){
    	score <-
    	utwoa(x,y,weights,cumlist[[1]],cumlist[[2]],
                          cumlist[[3]],cumlist[[4]])
    	infotwo <- effinfo(y,x,family,weights,psi,lambda,stepsize,nu)
    }
    else {
    	score <- uonea(x,y,weights,cumlist[[1]])
    	infotwo <- effinfouone(x,weights,cumlist[[2]])
    }
    teststat <- -sum((score %*% solve(infotwo) ) * score)
}

    results <- list(coef=coef,info=info,strata=rows,
    columns=cols,icnt=itcount,
    scoretest = paramtest,tstat = teststat, df = length(paramtest),
    call=call,weights =  
    weights, family = family)
    class(results) <- "psglm"
    return(results)
}

